# -*- coding: utf-8 -*-
from abjad.tools import datastructuretools
from abjad.tools import durationtools
from abjad.tools import mathtools
from abjad.tools import scoretools
from abjad.tools import selectiontools
from abjad.tools import sequencetools
from abjad.tools.rhythmmakertools.RhythmMaker import RhythmMaker
from abjad.tools.topleveltools import inspect_


class TupletRhythmMaker(RhythmMaker):
    r'''Tuplet rhythm-maker.

    ..  container:: example

        Makes tuplets with ``3:2`` ratios:

        ::

            >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
            ...     tuplet_ratios=[(3, 2)],
            ...     )

        ::

            >>> divisions = [(1, 2), (3, 8), (5, 16), (5, 16)]
            >>> selections = rhythm_maker(divisions)
            >>> lilypond_file = rhythmmakertools.make_lilypond_file(
            ...     selections,
            ...     divisions,
            ...     )
            >>> show(lilypond_file) # doctest: +SKIP

        ..  doctest::

            >>> f(lilypond_file[Staff])
            \new RhythmicStaff {
                {
                    \time 1/2
                    \times 4/5 {
                        c'4.
                        c'4
                    }
                }
                {
                    \time 3/8
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/5 {
                        c'4.
                        c'4
                    }
                }
                {
                    \time 5/16
                    {
                        c'8. [
                        c'8 ]
                    }
                }
                {
                    {
                        c'8. [
                        c'8 ]
                    }
                }
            }

    ..  container:: example

        Makes tuplets with alternating ``1:-1`` and ``3:1`` ratios:

        ::

            >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
            ...     tuplet_ratios=[(1, -1), (3, 1)],
            ...     )

        ::

            >>> divisions = [(1, 2), (3, 8), (5, 16), (5, 16)]
            >>> selections = rhythm_maker(divisions)
            >>> lilypond_file = rhythmmakertools.make_lilypond_file(
            ...     selections,
            ...     divisions,
            ...     )
            >>> show(lilypond_file) # doctest: +SKIP

        ..  doctest::

            >>> f(lilypond_file[Staff])
            \new RhythmicStaff {
                {
                    \time 1/2
                    {
                        c'4
                        r4
                    }
                }
                {
                    \time 3/8
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 3/4 {
                        c'4.
                        c'8
                    }
                }
                {
                    \time 5/16
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 5/6 {
                        c'8.
                        r8.
                    }
                }
                {
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 5/8 {
                        c'4.
                        c'8
                    }
                }
            }

    Object model of a partially evaluated function that accepts a (possibly
    empty) list of divisions as input and returns a list of selections as
    output. Output structured one selection per division with each selection
    wrapping a single fixed-duration tuplet.

    Usage follows the two-step configure-once / call-repeatedly pattern shown
    here.
    '''

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Rhythm-makers'

    __slots__ = (
        '_preferred_denominator',
        '_tuplet_ratios',
        '_tuplet_spelling_specifier',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        tuplet_ratios=None,
        beam_specifier=None,
        division_masks=None,
        duration_spelling_specifier=None,
        preferred_denominator=None,
        tie_specifier=None,
        tuplet_spelling_specifier=None,
        ):
        RhythmMaker.__init__(
            self,
            beam_specifier=beam_specifier,
            duration_spelling_specifier=duration_spelling_specifier,
            division_masks=division_masks,
            tie_specifier=tie_specifier,
            tuplet_spelling_specifier=tuplet_spelling_specifier,
            )
        if tuplet_ratios is not None:
            tuplet_ratios = tuple(mathtools.Ratio(x) for x in tuplet_ratios)
        self._tuplet_ratios = tuplet_ratios
        if preferred_denominator is not None:
            prototype = (durationtools.Duration, int)
            assert (preferred_denominator == 'divisions' or
                isinstance(preferred_denominator, prototype))
        self._preferred_denominator = preferred_denominator

    ### SPECIAL METHODS ###

    def __call__(self, divisions, rotation=None):
        r'''Calls tuplet rhythm-maker on `divisions`.

        ..  container:: example

            Calls tuplet rhythm-maker with one ratio:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(3, 2)],
                ...     )

            ::

                >>> divisions = [(1, 2), (3, 8), (5, 16), (5, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 1/2
                        \times 4/5 {
                            c'4.
                            c'4
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4.
                            c'4
                        }
                    }
                    {
                        \time 5/16
                        {
                            c'8. [
                            c'8 ]
                        }
                    }
                    {
                        {
                            c'8. [
                            c'8 ]
                        }
                    }
                }

        ..  container:: example

            Calls tuplet rhythm-maker on two ratios:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, -1), (3, 1)],
                ...     )

            ::

                >>> divisions = [(1, 2), (3, 8), (5, 16), (5, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 1/2
                        {
                            c'4
                            r4
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/4 {
                            c'4.
                            c'8
                        }
                    }
                    {
                        \time 5/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 5/6 {
                            c'8.
                            r8.
                        }
                    }
                    {
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 5/8 {
                            c'4.
                            c'8
                        }
                    }
                }

        Returns list of selections structured one selection per division.
        Each selection wraps a single fixed-duration tuplet.
        '''
        return RhythmMaker.__call__(
            self,
            divisions,
            rotation=rotation,
            )

    def __format__(self, format_specification=''):
        r'''Formats tuplet rhythm-maker.

        ..  container:: example

            Formats tuplet rhythm-maker with one ratio:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(3, 2)],
                ...     )

            ::

                >>> f(rhythm_maker)
                rhythmmakertools.TupletRhythmMaker(
                    tuplet_ratios=[
                        mathtools.Ratio((3, 2)),
                        ],
                    )

        ..  container:: example

            Formats tuplet rhythm-maker with two ratios:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, -1), (3, 1)],
                ...     )

            ::

                >>> f(rhythm_maker)
                rhythmmakertools.TupletRhythmMaker(
                    tuplet_ratios=[
                        mathtools.Ratio((1, -1)),
                        mathtools.Ratio((3, 1)),
                        ],
                    )

        Set `format_specification` to `''` or `'storage'`.

        Returns string.
        '''
        superclass = super(TupletRhythmMaker, self)
        return superclass.__format__(format_specification=format_specification)

    ### PRIVATE METHODS ###

    def _make_music(self, divisions, rotation):
        tuplets = []
        prototype = mathtools.NonreducedFraction
        assert all(isinstance(_, prototype) for _ in divisions)
        if not isinstance(rotation, int):
            rotation = 0
        tuplet_ratios = datastructuretools.CyclicTuple(
            sequencetools.Sequence(self.tuplet_ratios).rotate(n=rotation)
            )
        tuplet_spelling_specifier = self._get_tuplet_spelling_specifier()
        for duration_index, division in enumerate(divisions):
            ratio = tuplet_ratios[duration_index]
            duration = durationtools.Duration(division)
            tuplet = self._make_tuplet(
                duration,
                ratio,
                avoid_dots=tuplet_spelling_specifier.avoid_dots,
                is_diminution=tuplet_spelling_specifier.is_diminution,
                )
            preferred_denominator = \
                tuplet_spelling_specifier.preferred_denominator
            if preferred_denominator is None:
                pass
            elif preferred_denominator == 'divisions':
                tuplet.preferred_denominator = division.numerator
            elif isinstance(preferred_denominator, durationtools.Duration):
                unit_duration = preferred_denominator
                assert unit_duration.numerator == 1
                duration = inspect_(tuplet).get_duration()
                denominator = unit_duration.denominator
                nonreduced_fraction = duration.with_denominator(denominator)
                tuplet.preferred_denominator = nonreduced_fraction.numerator
            elif mathtools.is_positive_integer(preferred_denominator):
                tuplet.preferred_denominator = preferred_denominator
            else:
                raise ValueError(preferred_denominator)
            tuplets.append(tuplet)
        selections = [selectiontools.Selection(x) for x in tuplets]
        beam_specifier = self._get_beam_specifier()
        beam_specifier(selections)
        selections = self._apply_division_masks(selections, rotation)
        return selections

    def _make_tuplet(
        self,
        duration,
        ratio,
        avoid_dots=False,
        is_diminution=True,
        ):
        tuplet = scoretools.Tuplet.from_duration_and_ratio(
            duration,
            ratio,
            avoid_dots=avoid_dots,
            is_diminution=is_diminution,
            )
        return tuplet

    ### PUBLIC PROPERTIES ###

    @property
    def beam_specifier(self):
        r'''Gets beam specifier.

        ..  container:: example

            Beams each division:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 1, 2, 1, 1), (3, 1, 1)],
                ...     beam_specifier=rhythmmakertools.BeamSpecifier(
                ...         beam_each_division=True,
                ...         ),
                ...     )

            ::

                >>> divisions = [(5, 8), (3, 8), (6, 8), (4, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file)  # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 5/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 5/9 {
                            c'8. [
                            c'8. ]
                            c'4.
                            c'8. [
                            c'8. ]
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4.
                            c'8 [
                            c'8 ]
                        }
                    }
                    {
                        \time 6/8
                        {
                            c'8 [
                            c'8 ]
                            c'4
                            c'8 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/8
                        \times 4/5 {
                            c'4.
                            c'8 [
                            c'8 ]
                        }
                    }
                }

        ..  container:: example

            Beams divisions together:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 1, 2, 1, 1), (3, 1, 1)],
                ...     beam_specifier=rhythmmakertools.BeamSpecifier(
                ...         beam_divisions_together=True,
                ...         ),
                ...     )

            ::

                >>> divisions = [(5, 8), (3, 8), (6, 8), (4, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file)  # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 5/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 5/9 {
                            \set stemLeftBeamCount = #0
                            \set stemRightBeamCount = #1
                            c'8. [
                            \set stemLeftBeamCount = #1
                            \set stemRightBeamCount = #1
                            c'8. ]
                            c'4.
                            \set stemLeftBeamCount = #0
                            \set stemRightBeamCount = #1
                            c'8. [
                            \set stemLeftBeamCount = #1
                            \set stemRightBeamCount = #1
                            c'8. ]
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4.
                            \set stemLeftBeamCount = #0
                            \set stemRightBeamCount = #1
                            c'8 [
                            \set stemLeftBeamCount = #1
                            \set stemRightBeamCount = #1
                            c'8
                        }
                    }
                    {
                        \time 6/8
                        {
                            \set stemLeftBeamCount = #1
                            \set stemRightBeamCount = #1
                            c'8
                            \set stemLeftBeamCount = #1
                            \set stemRightBeamCount = #1
                            c'8 ]
                            c'4
                            \set stemLeftBeamCount = #0
                            \set stemRightBeamCount = #1
                            c'8 [
                            \set stemLeftBeamCount = #1
                            \set stemRightBeamCount = #1
                            c'8 ]
                        }
                    }
                    {
                        \time 4/8
                        \times 4/5 {
                            c'4.
                            \set stemLeftBeamCount = #0
                            \set stemRightBeamCount = #1
                            c'8 [
                            \set stemLeftBeamCount = #1
                            \set stemRightBeamCount = #0
                            c'8 ]
                        }
                    }
                }

        ..  container:: example

            Beams nothing:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 1, 2, 1, 1), (3, 1, 1)],
                ...     beam_specifier=rhythmmakertools.BeamSpecifier(
                ...         beam_divisions_together=False,
                ...         beam_each_division=False,
                ...         ),
                ...     )

            ::

                >>> divisions = [(5, 8), (3, 8), (6, 8), (4, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file)  # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 5/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 5/9 {
                            c'8.
                            c'8.
                            c'4.
                            c'8.
                            c'8.
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4.
                            c'8
                            c'8
                        }
                    }
                    {
                        \time 6/8
                        {
                            c'8
                            c'8
                            c'4
                            c'8
                            c'8
                        }
                    }
                    {
                        \time 4/8
                        \times 4/5 {
                            c'4.
                            c'8
                            c'8
                        }
                    }
                }

        Ignores `beam_each_division` when `beam_division_together` is true.

        Set to beam specifier or none.

        Returns beam specifier or none.
        '''
        superclass = super(TupletRhythmMaker, self)
        return superclass.beam_specifier

    @property
    def division_masks(self):
        r'''Gets division masks.

        ..  container:: example

            No division masks:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(4, 1)],
                ...     beam_specifier=rhythmmakertools.BeamSpecifier(
                ...         beam_divisions_together=False,
                ...         beam_each_division=False,
                ...         ),
                ...     )

            ::

                >>> divisions = [(3, 8), (4, 8), (3, 8), (4, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 3/8
                        \times 4/5 {
                            c'4.
                            c'16.
                        }
                    }
                    {
                        \time 4/8
                        \times 4/5 {
                            c'2
                            c'8
                        }
                    }
                    {
                        \time 3/8
                        \times 4/5 {
                            c'4.
                            c'16.
                        }
                    }
                    {
                        \time 4/8
                        \times 4/5 {
                            c'2
                            c'8
                        }
                    }
                }

        ..  container:: example

            Masks every other output division:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(4, 1)],
                ...     beam_specifier=rhythmmakertools.BeamSpecifier(
                ...         beam_divisions_together=False,
                ...         beam_each_division=False,
                ...         ),
                ...     division_masks=[
                ...         rhythmmakertools.silence_every([1], period=2),
                ...         ],
                ...     )

            ::

                >>> divisions = [(3, 8), (4, 8), (3, 8), (4, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 3/8
                        \times 4/5 {
                            c'4.
                            c'16.
                        }
                    }
                    {
                        \time 4/8
                        r2
                    }
                    {
                        \time 3/8
                        \times 4/5 {
                            c'4.
                            c'16.
                        }
                    }
                    {
                        \time 4/8
                        r2
                    }
                }

        Set to division masks or none.

        Returns division masks or none.
        '''
        superclass = super(TupletRhythmMaker, self)
        return superclass.division_masks

    @property
    def preferred_denominator(self):
        r'''Gets preferred denominator.

        ..  container:: example

            Tuplet numerators and denominators are reduced to numbers that are
            relatively prime when `preferred_denominator` is set to none. This
            means that ratios like ``6:4`` and ``10:8`` do not arise:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 4)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         preferred_denominator=None,
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 16), (4, 16), (6, 16), (8, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/16
                        \times 4/5 {
                            c'32 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/16
                        \times 4/5 {
                            c'16
                            c'4
                        }
                    }
                    {
                        \time 6/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'8
                            c'2
                        }
                    }
                    {
                        \time 8/16
                        \times 4/5 {
                            c'8
                            c'2
                        }
                    }
                }

        ..  container:: example

            The preferred denominator of each tuplet is set to the numerator of
            the division that generates the tuplet when `preferred_denominator`
            is set to the string ``'divisions'``. This means that the tuplet
            numerator and denominator are not necessarily relatively prime.
            This also means that ratios like ``6:4`` and ``10:8`` may arise:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 4)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         preferred_denominator='divisions',
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 16), (4, 16), (6, 16), (8, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/16
                        \times 4/5 {
                            c'32 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/16
                        \times 4/5 {
                            c'16
                            c'4
                        }
                    }
                    {
                        \time 6/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 6/10 {
                            c'8
                            c'2
                        }
                    }
                    {
                        \time 8/16
                        \times 8/10 {
                            c'8
                            c'2
                        }
                    }
                }

        ..  container:: example

            The preferred denominator of each tuplet is set in terms of a unit
            duration when `preferred_denominator` is set to a duration. The
            setting does not affect the first tuplet:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 4)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         preferred_denominator=Duration(1, 16),
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 16), (4, 16), (6, 16), (8, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/16
                        \times 4/5 {
                            c'32 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/16
                        \times 4/5 {
                            c'16
                            c'4
                        }
                    }
                    {
                        \time 6/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 6/10 {
                            c'8
                            c'2
                        }
                    }
                    {
                        \time 8/16
                        \times 8/10 {
                            c'8
                            c'2
                        }
                    }
                }

        ..  container:: example

            Sets the preferred denominator of each tuplet in terms 32nd notes.
            The setting affects all tuplets:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 4)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         preferred_denominator=Duration(1, 32),
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 16), (4, 16), (6, 16), (8, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/16
                        \times 4/5 {
                            c'32 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/16
                        \times 8/10 {
                            c'16
                            c'4
                        }
                    }
                    {
                        \time 6/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 12/20 {
                            c'8
                            c'2
                        }
                    }
                    {
                        \time 8/16
                        \times 16/20 {
                            c'8
                            c'2
                        }
                    }
                }

        ..  container:: example

            Sets the preferred denominator each tuplet in terms 64th notes. The
            setting affects all tuplets:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 4)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         preferred_denominator=Duration(1, 64),
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 16), (4, 16), (6, 16), (8, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/16
                        \times 8/10 {
                            c'32 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/16
                        \times 16/20 {
                            c'16
                            c'4
                        }
                    }
                    {
                        \time 6/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 24/40 {
                            c'8
                            c'2
                        }
                    }
                    {
                        \time 8/16
                        \times 32/40 {
                            c'8
                            c'2
                        }
                    }
                }

        ..  container:: example

            The preferred denominator of each tuplet is set directly when
            `preferred_denominator` is set to a positive integer. This example
            sets the preferred denominator of each tuplet to ``8``. Setting
            does not affect the third tuplet:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 4)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         preferred_denominator=8,
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 16), (4, 16), (6, 16), (8, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/16
                        \times 8/10 {
                            c'32 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/16
                        \times 8/10 {
                            c'16
                            c'4
                        }
                    }
                    {
                        \time 6/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'8
                            c'2
                        }
                    }
                    {
                        \time 8/16
                        \times 8/10 {
                            c'8
                            c'2
                        }
                    }
                }

        ..  container:: example

            Sets the preferred denominator of each tuplet to ``12``. Setting
            affects all tuplets:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 4)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         preferred_denominator=12,
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 16), (4, 16), (6, 16), (8, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/16
                        \times 12/15 {
                            c'32 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/16
                        \times 12/15 {
                            c'16
                            c'4
                        }
                    }
                    {
                        \time 6/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 12/20 {
                            c'8
                            c'2
                        }
                    }
                    {
                        \time 8/16
                        \times 12/15 {
                            c'8
                            c'2
                        }
                    }
                }

        ..  container:: example

            Sets the preferred denominator of each tuplet to ``13``. Setting
            does not affect any tuplet:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 4)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         preferred_denominator=13,
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 16), (4, 16), (6, 16), (8, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/16
                        \times 4/5 {
                            c'32 [
                            c'8 ]
                        }
                    }
                    {
                        \time 4/16
                        \times 4/5 {
                            c'16
                            c'4
                        }
                    }
                    {
                        \time 6/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'8
                            c'2
                        }
                    }
                    {
                        \time 8/16
                        \times 4/5 {
                            c'8
                            c'2
                        }
                    }
                }

        Set to ``'divisions'``, duration, positive integer or none.

        Returns ``'divisions'``, duration, positive integer or none.
        '''
        return self._preferred_denominator

    @property
    def tie_specifier(self):
        r'''Gets tie specifier.

        ..  container:: example

            Ties nothing:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(2, 3), (1, -2, 1)],
                ...     tie_specifier=rhythmmakertools.TieSpecifier(
                ...         tie_across_divisions=False,
                ...         ),
                ...     )

            ::

                >>> divisions = [(1, 2), (3, 8), (5, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 1/2
                        \times 4/5 {
                            c'4
                            c'4.
                        }
                    }
                    {
                        \time 3/8
                        {
                            c'16.
                            r8.
                            c'16.
                        }
                    }
                    {
                        \time 5/16
                        {
                            c'8 [
                            c'8. ]
                        }
                    }
                }

        ..  container:: example

            Ties across all divisions:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(2, 3), (1, -2, 1)],
                ...     tie_specifier=rhythmmakertools.TieSpecifier(
                ...         tie_across_divisions=True,
                ...         ),
                ...     )

            ::

                >>> divisions = [(1, 2), (3, 8), (5, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 1/2
                        \times 4/5 {
                            c'4
                            c'4. ~
                        }
                    }
                    {
                        \time 3/8
                        {
                            c'16.
                            r8.
                            c'16. ~
                        }
                    }
                    {
                        \time 5/16
                        {
                            c'8 [
                            c'8. ]
                        }
                    }
                }

        ..  container:: example

            Ties across every other division:

            ::

                >>> pattern = patterntools.Pattern(
                ...     indices=[0],
                ...     period=2,
                ...     )
                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(2, 3), (1, -2, 1)],
                ...     tie_specifier=rhythmmakertools.TieSpecifier(
                ...         tie_across_divisions=pattern,
                ...         ),
                ...     )

            ::

                >>> divisions = [(1, 2), (3, 8), (5, 16), (5, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 1/2
                        \times 4/5 {
                            c'4
                            c'4. ~
                        }
                    }
                    {
                        \time 3/8
                        {
                            c'16.
                            r8.
                            c'16.
                        }
                    }
                    {
                        \time 5/16
                        {
                            c'8 [
                            c'8. ~ ]
                        }
                    }
                    {
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 5/6 {
                            c'16.
                            r8.
                            c'16.
                        }
                    }
                }

        Set to tie specifier or none.

        Returns tie speicifer or none.
        '''
        return RhythmMaker.tie_specifier.fget(self)

    @property
    def tuplet_ratios(self):
        r'''Gets tuplet ratios.

        ..  container:: example

            Makes tuplets with ``3:2`` ratios:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(3, 2)],
                ...     )

            ::

                >>> divisions = [(1, 2), (3, 8), (5, 16), (5, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 1/2
                        \times 4/5 {
                            c'4.
                            c'4
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4.
                            c'4
                        }
                    }
                    {
                        \time 5/16
                        {
                            c'8. [
                            c'8 ]
                        }
                    }
                    {
                        {
                            c'8. [
                            c'8 ]
                        }
                    }
                }

        ..  container:: example

            Makes tuplets with alternating ``1:-1`` and ``3:1`` ratios:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, -1), (3, 1)],
                ...     )

            ::

                >>> divisions = [(1, 2), (3, 8), (5, 16), (5, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 1/2
                        {
                            c'4
                            r4
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/4 {
                            c'4.
                            c'8
                        }
                    }
                    {
                        \time 5/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 5/6 {
                            c'8.
                            r8.
                        }
                    }
                    {
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 5/8 {
                            c'4.
                            c'8
                        }
                    }
                }

        Set to tuple of ratios.

        Returns tuple of ratios.
        '''
        if self._tuplet_ratios:
            return list(self._tuplet_ratios)

    @property
    def tuplet_spelling_specifier(self):
        r'''Gets tuplet spelling specifier.

        ..  container:: example

            Makes diminished tuplets and does not avoid dots:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 1)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=False,
                ...         is_diminution=True,
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 8), (3, 8), (7, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/8
                        {
                            c'8 [
                            c'8 ]
                        }
                    }
                    {
                        \time 3/8
                        {
                            c'8. [
                            c'8. ]
                        }
                    }
                    {
                        \time 7/16
                        {
                            c'8.. [
                            c'8.. ]
                        }
                    }
                }

        ..  container:: example

            Makes diminished tuplets and avoids dots:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 1)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         is_diminution=True,
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 8), (3, 8), (7, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/8
                        {
                            c'8 [
                            c'8 ]
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/4 {
                            c'4
                            c'4
                        }
                    }
                    {
                        \time 7/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 7/8 {
                            c'4
                            c'4
                        }
                    }
                }

        ..  container:: example

            Makes augmented tuplets and does not avoid dots:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 1)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=False,
                ...         is_diminution=False,
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 8), (3, 8), (7, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/8
                        {
                            c'8 [
                            c'8 ]
                        }
                    }
                    {
                        \time 3/8
                        {
                            c'8. [
                            c'8. ]
                        }
                    }
                    {
                        \time 7/16
                        {
                            c'8.. [
                            c'8.. ]
                        }
                    }
                }

        ..  container:: example

            Makes augmented tuplets and avoids dots:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(1, 1)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         is_diminution=False,
                ...         ),
                ...     )

            ::

                >>> divisions = [(2, 8), (3, 8), (7, 16)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 2/8
                        {
                            c'8 [
                            c'8 ]
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/2 {
                            c'8 [
                            c'8 ]
                        }
                    }
                    {
                        \time 7/16
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 7/4 {
                            c'8 [
                            c'8 ]
                        }
                    }
                }

        ..  container:: example

            Does not simplify tuplets:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(3, -2), (1,), (-2, 3), (1, 1)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         simplify_redundant_tuplets=False,
                ...         ),
                ...     )

            ::

                >>> divisions = [(3, 8), (3, 8), (3, 8), (3, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4.
                            r4
                        }
                    }
                    {
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/4 {
                            c'2
                        }
                    }
                    {
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            r4
                            c'4.
                        }
                    }
                    {
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/4 {
                            c'4
                            c'4
                        }
                    }
                }

        ..  container:: example

            Simplifies tuplets:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tuplet_ratios=[(3, -2), (1,), (-2, 3), (1, 1)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         avoid_dots=True,
                ...         simplify_redundant_tuplets=True,
                ...         ),
                ...     )

            ::

                >>> divisions = [(3, 8), (3, 8), (3, 8), (3, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ..  doctest::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4.
                            r4
                        }
                    }
                    {
                        {
                            c'4.
                        }
                    }
                    {
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            r4
                            c'4.
                        }
                    }
                    {
                        {
                            c'8. [
                            c'8. ]
                        }
                    }
                }

        ..  container:: example

            Leaves trivial tuplets enclosed in curly braces in LilyPond output:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tie_specifier=rhythmmakertools.TieSpecifier(
                ...         tie_across_divisions=True,
                ...         ),
                ...     tuplet_ratios=[(2, 3), (1, 1)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         flatten_trivial_tuplets=False,
                ...         ),
                ...     )

            ::

                >>> divisions = [(3, 8), (2, 8), (3, 8), (2, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4
                            c'4. ~
                        }
                    }
                    {
                        \time 2/8
                        {
                            c'8 [
                            c'8 ~ ]
                        }
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4
                            c'4. ~
                        }
                    }
                    {
                        \time 2/8
                        {
                            c'8 [
                            c'8 ]
                        }
                    }
                }

            Runs of eighth notes are enclosed in a first set of curly braces
            (representing trivial tuplets) and a second set of curly braces
            (representing measures).

        ..  container:: example

            Flattens trivial tuplets:

            ::

                >>> rhythm_maker = rhythmmakertools.TupletRhythmMaker(
                ...     tie_specifier=rhythmmakertools.TieSpecifier(
                ...         tie_across_divisions=True,
                ...         ),
                ...     tuplet_ratios=[(2, 3), (1, 1)],
                ...     tuplet_spelling_specifier=rhythmmakertools.TupletSpellingSpecifier(
                ...         flatten_trivial_tuplets=True,
                ...         ),
                ...     )

            ::

                >>> divisions = [(3, 8), (2, 8), (3, 8), (2, 8)]
                >>> selections = rhythm_maker(divisions)
                >>> lilypond_file = rhythmmakertools.make_lilypond_file(
                ...     selections,
                ...     divisions,
                ...     )
                >>> show(lilypond_file) # doctest: +SKIP

            ::

                >>> f(lilypond_file[Staff])
                \new RhythmicStaff {
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4
                            c'4. ~
                        }
                    }
                    {
                        \time 2/8
                        c'8 [
                        c'8 ~ ]
                    }
                    {
                        \time 3/8
                        \tweak text #tuplet-number::calc-fraction-text
                        \times 3/5 {
                            c'4
                            c'4. ~
                        }
                    }
                    {
                        \time 2/8
                        c'8 [
                        c'8 ]
                    }
                }

            Runs of eighth notes are now enclosed in only one set of curly
            braces (representing measures). The graphic output of the two
            examples is the same.

            .. note:: Flattening trivial tuplets makes it possible 
                subsequently to rewrite the meter of the untupletted notes.

        Defaults to none.

        Set to tuplet spelling specifier or none.

        Returns tuplet spelling specifier or none.
        '''
        superclass = super(TupletRhythmMaker, self)
        return superclass.tuplet_spelling_specifier



# -*- coding: utf-8 -*-
import multiprocessing
import pickle
from abjad.tools.quantizationtools.JobHandler import JobHandler


class ParallelJobHandler(JobHandler):
    r'''Processes ``QuantizationJob`` instances in parallel,
    based on the number of CPUs available.
    '''

    ### SPECIAL METHODS ###

    def __call__(self, jobs):
        r'''Calls parallel job handler.
        '''
        from abjad.tools import quantizationtools
        finished_jobs = []
        job_queue = multiprocessing.JoinableQueue()
        result_queue = multiprocessing.Queue()
        workers = [quantizationtools.ParallelJobHandlerWorker(
            job_queue, result_queue)
            for i in range(multiprocessing.cpu_count() * 2)]
        for worker in workers:
            worker.start( )
        for job in jobs:
            job_queue.put(pickle.dumps(job, protocol=0))
        for i in range(len(jobs)):
            finished_jobs.append(pickle.loads(result_queue.get()))
        for worker in workers:
            job_queue.put(None)
        job_queue.join()
        result_queue.close()
        job_queue.close()
        for worker in workers:
             worker.join()
        return finished_jobs



# -*- coding: utf-8 -*-
from abjad import *


def test_spannertools_TextSpanner___init___01():
    r'''Initialize empty text spanner.
    '''

    spanner = spannertools.TextSpanner()
    assert isinstance(spanner, spannertools.TextSpanner)



# -*- coding: utf-8 -*-
from abjad import *


def test_pitchtools_PitchClassSet_order_by_01():

    named_pitch_class_set = pitchtools.PitchClassSet(['c', 'e', 'b'])
    named_pitch_class_segment = pitchtools.PitchClassSegment(['e', 'a', 'f'])
    ordered_set = named_pitch_class_set.order_by(
        named_pitch_class_segment)

    assert ordered_set == pitchtools.PitchClassSegment(['b', 'e', 'c'])



# -*- coding: utf-8 -*-
from abjad.tools.abctools import AbjadObject


class LilyPondDuration(AbjadObject):
    r'''Model of a duration in LilyPond.

    Not composer-safe.

    Used internally by LilyPondParser.
    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        'duration',
        'multiplier',
        )

    ### INITIALIZER ###

    def __init__(self, duration=None, multiplier=None):
        self.duration = duration
        self.multiplier = multiplier



# -*- coding: utf-8 -*-
from abjad import *
import pytest


def test_scoretools_Measure_scale_and_adjust_time_signature_01():
    r'''Scales power-of-two measure to non-power-of-two measure.
    No note-head rewriting necessary.
    '''

    measure = Measure((3, 8), "c'8 d'8 e'8")
    measure.scale_and_adjust_time_signature(Multiplier(2, 3))

    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 3/12
            \scaleDurations #'(2 . 3) {
                c'8
                d'8
                e'8
            }
        }
        '''
        )

    assert inspect_(measure).is_well_formed()



def test_scoretools_Measure_scale_and_adjust_time_signature_02():
    r'''Scale non-power-of-two time signature to power-of-two.
    No note-head rewriting necessary.
    '''

    measure = Measure((3, 12), "c'8 d'8 e'8")
    measure.scale_and_adjust_time_signature(Multiplier(3, 2))

    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 3/8
            c'8
            d'8
            e'8
        }
        '''
        )

    assert inspect_(measure).is_well_formed()


def test_scoretools_Measure_scale_and_adjust_time_signature_03():
    r'''Scale power-of-two time signature to power-of-two time signature.
    Noteheads rewrite with dots.
    '''

    measure = Measure((3, 8), "c'8 d'8 e'8")
    measure.scale_and_adjust_time_signature(Multiplier(3, 2))

    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 9/16
            c'8.
            d'8.
            e'8.
        }
        '''
        )

    assert inspect_(measure).is_well_formed()


def test_scoretools_Measure_scale_and_adjust_time_signature_04():
    r'''Scale power-of-two time signature to power-of-two time signature.
    Noteheads rewrite without dots.
    '''

    measure = Measure((9, 16), "c'8. d'8. e'8.")
    measure.scale_and_adjust_time_signature(Multiplier(2, 3))

    assert inspect_(measure).is_well_formed()
    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 3/8
            c'8
            d'8
            e'8
        }
        '''
        )


def test_scoretools_Measure_scale_and_adjust_time_signature_05():
    r'''Scale power-of-two time signature to non-power-of-two time signature.
    No note-head rewriting necessary.
    '''

    measure = Measure((9, 16), "c'16 d'16 e'16 f'16 g'16 a'16 b'16 c''16 d''16")
    measure.scale_and_adjust_time_signature(Multiplier(2, 3))

    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 9/24
            \scaleDurations #'(2 . 3) {
                c'16
                d'16
                e'16
                f'16
                g'16
                a'16
                b'16
                c''16
                d''16
            }
        }
        '''
        )

    assert inspect_(measure).is_well_formed()


def test_scoretools_Measure_scale_and_adjust_time_signature_06():
    r'''Scale non-power-of-two time signature to power-of-two time signature.
    Noteheads rewrite with double duration.
    '''

    measure = Measure((3, 12), "c'8 d'8 e'8")
    measure.scale_and_adjust_time_signature(Multiplier(3))

    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 3/4
            c'4
            d'4
            e'4
        }
        '''
        )

    assert inspect_(measure).is_well_formed()


def test_scoretools_Measure_scale_and_adjust_time_signature_07():
    r'''Scale power-of-two time signature by one half.
    Noteheads rewrite with half duration.
    Time signature rewrites with double denominator.
    '''

    measure = Measure((6, 16), "c'16 d'16 e'16 f'16 g'16 a'16")
    measure.scale_and_adjust_time_signature(Multiplier(1, 2))

    assert inspect_(measure).is_well_formed()
    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 6/32
            c'32
            d'32
            e'32
            f'32
            g'32
            a'32
        }
        '''
        )


def test_scoretools_Measure_scale_and_adjust_time_signature_08():
    r'''Scale power-of-two time signature by one quarter.
    Noteheads rewrite with quarter duration.
    Time signature rewrites with quadruple denominator.
    '''

    measure = Measure((6, 16), "c'16 d'16 e'16 f'16 g'16 a'16")
    measure.scale_and_adjust_time_signature(Multiplier(1, 4))

    assert inspect_(measure).is_well_formed()
    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 6/64
            c'64
            d'64
            e'64
            f'64
            g'64
            a'64
        }
        '''
        )


def test_scoretools_Measure_scale_and_adjust_time_signature_09():
    r'''Scale power-of-two time signature by two.
    Noteheads rewrite with double duration.
    Time signature rewrites with half denominator.
    '''

    measure = Measure((6, 16), "c'16 d'16 e'16 f'16 g'16 a'16")
    measure.scale_and_adjust_time_signature(Multiplier(2))

    assert inspect_(measure).is_well_formed()
    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 6/8
            c'8
            d'8
            e'8
            f'8
            g'8
            a'8
        }
        '''
        )


def test_scoretools_Measure_scale_and_adjust_time_signature_10():
    r'''Scale power-of-two time signature by four.
    Noteheads rewrite with quadruple duration.
    Time signature rewrites with quarter denominator.
    '''

    measure = Measure((6, 16), "c'16 d'16 e'16 f'16 g'16 a'16")
    measure.scale_and_adjust_time_signature(Multiplier(4))

    assert inspect_(measure).is_well_formed()
    assert format(measure) == stringtools.normalize(
        r'''
        {
            \time 6/4
            c'4
            d'4
            e'4
            f'4
            g'4
            a'4
        }
        '''
        )



# -*- coding: utf-8 -*-
import abc
import math
import re
from abjad.tools import mathtools
from abjad.tools import systemtools
from abjad.tools.abctools import AbjadValueObject
from abjad.tools.pitchtools.Accidental import Accidental
from abjad.tools.pitchtools.Octave import Octave
from abjad.tools.pitchtools.PitchClass import PitchClass


class Pitch(AbjadValueObject):
    '''Pitch base class.
    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        )

    _diatonic_pitch_name_regex_body = '''
        {}  # exactly one diatonic pitch-class name
        {}  # followed by exactly one octave tick string
        '''.format(
        PitchClass._diatonic_pitch_class_name_regex_body,
        Octave._octave_tick_regex_body,
        )

    _diatonic_pitch_name_regex = re.compile(
        '^{}$'.format(_diatonic_pitch_name_regex_body),
        re.VERBOSE,
        )

    _pitch_class_octave_number_regex_body = '''
        (
        (?P<diatonic_pitch_class_name>
            [A-G]   # exactly one diatonic pitch-class name
        )
        {}          # plus an optional symbolic accidental string
        (?P<octave_number>
            [-]?    # plus an optional negative sign
            [0-9]+  # plus one or more digits
        )
        )
        '''.format(
        Accidental._symbolic_string_regex_body,
        )

    _pitch_class_octave_number_regex = re.compile(
        '^{}$'.format(_pitch_class_octave_number_regex_body),
        re.VERBOSE,
        )

    _pitch_name_regex_body = '''
        {}  # exactly one diatonic pitch-class name
        {}  # followed by exactly one alphabetic accidental name
        {}  # followed by exactly one octave tick string
        '''.format(
        PitchClass._diatonic_pitch_class_name_regex_body,
        Accidental._alphabetic_accidental_regex_body,
        Octave._octave_tick_regex_body,
        )

    _pitch_name_regex = re.compile(
        '^{}$'.format(_pitch_name_regex_body),
        re.VERBOSE,
        )

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self):
        pass

    ### SPECIAL METHODS ###

    @abc.abstractmethod
    def __float__(self):
        r'''Changes pitch to float.

        Returns float.
        '''
        message = 'TODO: all pitch-related classes must implement float.'
        raise NotImplementedError(message)

    def __format__(self, format_specification=''):
        r'''Formats pitch.

        Set `format_specification` to `''`, `'lilypond'` or `'storage'`.

        Returns string.
        '''
        from abjad.tools import systemtools
        if format_specification in ('', 'lilypond'):
            return self._get_lilypond_format()
        elif format_specification == 'storage':
            return systemtools.StorageFormatAgent(self).get_storage_format()
        return str(self)

    def __illustrate__(self):
        r'''Illustrates pitch.

        Returns LilyPond file.
        '''
        import abjad
        pitch = abjad.NamedPitch(self)
        note = abjad.Note(pitch, 1)
        abjad.attach(abjad.Multiplier(1, 4), note)
        clef = abjad.Clef.from_selection([pitch])
        staff = abjad.Staff()
        abjad.attach(clef, staff)
        staff.append(note)
        abjad.override(staff).time_signature.stencil = False
        lilypond_file = abjad.lilypondfiletools.LilyPondFile.new(staff)
        lilypond_file.header_block.tagline = False
        return lilypond_file

    @abc.abstractmethod
    def __int__(self):
        r'''Changes pitch to integer.

        Returns integer.
        '''
        message = 'TODO: all pitch-related classes must implement int.'
        raise NotImplementedError(message)

    ### PRIVATE PROPERTIES ###

    def _get_format_specification(self):
        if type(self).__name__.startswith('Named'):
            values = [str(self)]
        else:
            values = [
                mathtools.integer_equivalent_number_to_integer(float(self))
                ]
        return systemtools.FormatSpecification(
            client=self,
            coerce_for_equality=True,
            repr_is_indented=False,
            storage_format_is_indented=False,
            storage_format_args_values=values,
            template_names=['pitch_name'],
            )

    @abc.abstractmethod
    def _get_lilypond_format(self):
        raise NotImplementedError

    @staticmethod
    def _to_pitch_class_item_class(item_class):
        import abjad
        item_class = item_class or abjad.NumberedPitch
        if item_class in (abjad.NamedPitchClass, abjad.NumberedPitchClass):
            return item_class
        elif item_class is abjad.NamedPitch:
            return abjad.NamedPitchClass
        elif item_class is abjad.NumberedPitch:
            return abjad.NumberedPitchClass
        else:
            raise TypeError(item_class)

    @staticmethod
    def _to_pitch_item_class(item_class):
        import abjad
        item_class = item_class or abjad.NumberedPitch
        if item_class in (abjad.NamedPitch, abjad.NumberedPitch):
            return item_class
        elif item_class is abjad.NamedPitchClass:
            return abjad.NamedPitch
        elif item_class is abjad.NumberedPitchClass:
            return abjad.NumberedPitch
        else:
            raise TypeError(item_class)

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def alteration_in_semitones(self):
        r'''Gets alteration of pitch in semitones.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def diatonic_pitch_class_name(self):
        r'''Gets diatonic pitch-class name of pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def diatonic_pitch_class_number(self):
        r'''Gets diatonic pitch-class number of pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def diatonic_pitch_name(self):
        r'''Gets diatonic pitch name of pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def diatonic_pitch_number(self):
        r'''Gets diatonic pitch number of pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @property
    def hertz(self):
        r'''Gets frequency of pitch in Herz.

        ..  container:: example

            ::

                >>> pitchtools.NamedPitch("a'").hertz
                440.0

            ::

                >>> pitchtools.NamedPitch("c'").hertz
                261.62...

            ::

                >>> pitchtools.NamedPitch("c''").hertz
                523.25...

        Returns float.
        '''
        hertz = pow(2., (float(self) - 9.) / 12.) * 440.
        return hertz

    @abc.abstractproperty
    def name(self):
        r'''Gets pitch name.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def named_pitch(self):
        r'''Gets named pitch corresponding to pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def named_pitch_class(self):
        r'''Gets named pitch-class corresponding to pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def number(self):
        r'''Gets pitch number.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def numbered_pitch(self):
        r'''Gets numbered pitch corresponding to pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def numbered_pitch_class(self):
        r'''Gets numbered pitch-class corresponding to pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def octave(self):
        r'''Gets octave of pitch.

        Returns octave.
        '''
        raise NotImplementedError
            
    @abc.abstractproperty
    def pitch_class(self):
        r'''Gets pitch-class.

        Returns pitch-class.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def pitch_class_name(self):
        r'''Gets pitch-class name corresponding to pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def pitch_class_number(self):
        r'''Gets pitch-class number of pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def pitch_class_octave_label(self):
        r'''Gets pitch-class / octave label of pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def pitch_name(self):
        r'''Gets pitch name of pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    @abc.abstractproperty
    def pitch_number(self):
        r'''Get pitch number of pitch.

        ..  note:: Deprecated.
        '''
        raise NotImplementedError

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def apply_accidental(self, accidental=None):
        r'''Applies `accidental` to pitch.

        Returns new pitch.
        '''
        raise NotImplementedError

    @classmethod
    def from_hertz(class_, hertz):
        r'''Creates pitch from `hertz`.

        ..  container:: example

            ::

                >>> pitchtools.NamedPitch.from_hertz(440)
                NamedPitch("a'")

            ::

                >>> pitchtools.NumberedPitch.from_hertz(440)
                NumberedPitch(9)

            ::

                >>> pitchtools.NamedPitch.from_hertz(519)
                NamedPitch("c'")

        Returns new pitch.
        '''
        hertz = float(hertz)
        midi = 9. + (12. * math.log(hertz / 440., 2))
        pitch = class_(midi)
        return pitch

    @abc.abstractmethod
    def invert(self, axis=None):
        r'''Inverts pitch about `axis`.

        Interprets `axis` of none equal to middle C.

        Returns new pitch.
        '''
        from abjad.tools import pitchtools
        axis = axis or pitchtools.NamedPitch("c'")
        axis = type(self)(axis)
        interval = self - axis
        result = axis.transpose(interval)
        return result

    @staticmethod
    def is_diatonic_pitch_name(argument):
        '''Is true when `argument` is a diatonic pitch name. Otherwise false.

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_diatonic_pitch_name("c''")
                True

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_diatonic_pitch_name("cs''")
                False

        The regex ``(^[a-g,A-G])(,+|'+|)$`` underlies this predicate.

        Returns true or false.
        '''
        if not isinstance(argument, str):
            return False
        return bool(Pitch._diatonic_pitch_name_regex.match(argument))

    @staticmethod
    def is_diatonic_pitch_number(argument):
        '''Is true when `argument` is a diatonic pitch number. Otherwise false.

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_diatonic_pitch_number(7)
                True

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_diatonic_pitch_number(7.5)
                False

        Diatonic pitch numbers are equal to the set of integers.

        Returns true or false.
        '''
        return isinstance(argument, int)

    @staticmethod
    def is_pitch_carrier(argument):
        '''Is true when `argument` is an Abjad pitch, note, note-head of chord
        instance. Otherwise false.

        ..  container:: example

            ::

                >>> note = Note("c'4")
                >>> pitchtools.Pitch.is_pitch_carrier(note)
                True

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_pitch_carrier('text')
                False

        Returns true or false.
        '''
        from abjad.tools import scoretools
        from abjad.tools import pitchtools
        return isinstance(
            argument, (
                pitchtools.NamedPitch,
                scoretools.Note,
                scoretools.NoteHead,
                scoretools.Chord
                )
            )

    @staticmethod
    def is_pitch_class_octave_number_string(argument):
        '''Is true when `argument` is a pitch-class / octave number string.
        Otherwise false.

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_pitch_class_octave_number_string('C#2')
                True

        ..  container:: example

            Supports quartertone accidentals:

            ::

                >>> pitchtools.Pitch.is_pitch_class_octave_number_string('C#2')
                True

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_pitch_class_octave_number_string('C#')
                False

        The regex ``^([A-G])([#]{1,2}|[b]{1,2}|[#]?[+]|[b]?[~]|)([-]?[0-9]+)$``
        underlies this predicate.

        Returns true or false.
        '''
        if not isinstance(argument, str):
            return False
        return bool(Pitch._pitch_class_octave_number_regex.match(argument))

    @staticmethod
    def is_pitch_name(argument):
        '''Is true when `argument` is a pitch name. Otherwise false.

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_pitch_name('c,')
                True

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_pitch_name('z')
                False

        The regex ``^([a-g,A-G])(([s]{1,2}|[f]{1,2}|t?q?[f,s]|)!?)(,+|'+|)$``
        underlies this predicate.

        Returns true or false.
        '''
        if not isinstance(argument, str):
            return False
        return bool(Pitch._pitch_name_regex.match(argument))

    @staticmethod
    def is_pitch_number(argument):
        '''Is true when `argument` is a pitch number. Otherwise false.

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_pitch_number(13)
                True

        ..  container:: example

            ::

                >>> pitchtools.Pitch.is_pitch_number('text')
                False

        Pitch numbers are equal to the set of all integers in union with the
        set of all integers plus of minus 0.5.

        Returns true or false.
        '''
        if isinstance(argument, (int, float)):
            return argument % 0.5 == 0
        return False

    @abc.abstractmethod
    def multiply(self, n=1):
        r'''Multiplies pitch by `n`.
        '''
        raise NotImplementedError

    @abc.abstractmethod
    def transpose(self, n):
        r'''Transposes pitch by index `n`.

        Returns new pitch.
        '''
        raise NotImplementedError



# -*- coding: utf-8 -*-
from abjad.tools import sequencetools
from abjad.tools.pitchtools.Set import Set


class IntervalSet(Set):
    r'''Interval set.
    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        )

    ### INITIALIZER ###

    def __init__(self, items=None, item_class=None):
        from abjad.tools import pitchtools
        prototype = (
            pitchtools.PitchClassSegment,
            pitchtools.PitchClassSet,
            pitchtools.PitchSegment,
            pitchtools.PitchSet,
            )
        if isinstance(items, prototype):
            items = list(items)
            enumeration = sequencetools.Enumeration(items)
            pairs = enumeration.yield_pairs()
            items = [second - first for first, second in pairs]
        Set.__init__(
            self,
            items=items,
            item_class=item_class,
            )

    ### PRIVATE PROPERTIES ###

    @property
    def _named_item_class(self):
        from abjad.tools import pitchtools
        return pitchtools.NamedInterval

    @property
    def _numbered_item_class(self):
        from abjad.tools import pitchtools
        return pitchtools.NumberedInterval

    @property
    def _parent_item_class(self):
        from abjad.tools import pitchtools
        return pitchtools.Interval

    ### PUBLIC METHODS ###

    @classmethod
    def from_selection(class_, selection, item_class=None):
        r'''Initialize interval set from component selection:

        ::

            >>> staff_1 = Staff("c'4 <d' fs' a'>4 b2")
            >>> staff_2 = Staff("c4. r8 g2")
            >>> selection = select((staff_1, staff_2))
            >>> intervals = pitchtools.IntervalSet.from_selection(
            ...     selection)
            >>> for interval in sorted(intervals):
            ...     interval
            ...
            NamedInterval('-M6')
            NamedInterval('-P5')
            NamedInterval('-aug4')
            NamedInterval('-M3')
            NamedInterval('-m3')
            NamedInterval('-M2')
            NamedInterval('+m2')
            NamedInterval('+m3')
            NamedInterval('+M3')
            NamedInterval('+P4')
            NamedInterval('+P5')
            NamedInterval('+m7')
            NamedInterval('+M7')
            NamedInterval('+P8')
            NamedInterval('+M9')
            NamedInterval('+aug11')
            NamedInterval('+M13')

        Returns interval set.
        '''
        from abjad.tools import pitchtools
        pitch_segment = pitchtools.PitchSegment.from_selection(selection)
        enumeration = sequencetools.Enumeration(pitch_segment)
        pairs = enumeration.yield_pairs()
        intervals = (second - first for first, second in pairs)
        return class_(
            items=intervals,
            item_class=item_class,
            )



# -*- coding: utf-8 -*-
from abjad import *


def test_scoretools_NoteHead_is_forced_01():

    note_head = scoretools.NoteHead(written_pitch="c'")
    assert note_head.is_forced is None
    note_head.is_forced = True
    assert note_head.is_forced == True
    note_head.is_forced = False
    assert note_head.is_forced == False



# -*- coding: utf-8 -*-


_lilypond_parsers_by_language = {}


def parse(string, language='english'):
    r'''Parses LilyPond `string`.

    ..  container:: example

        Parses LilyPond string with English note names:

        ::

            >>> container = parse("{c'4 d'4 e'4 f'4}")
            >>> show(container) # doctest: +SKIP

    ..  container:: example

        Parses LilyPond string with Dutch note names:

        ::

            >>> container = parse("{c'8 des' e' fis'}", language='nederlands')
            >>> show(container) # doctest: +SKIP

    Returns Abjad component.
    '''
    from abjad.tools import rhythmtreetools
    from abjad.tools import lilypondparsertools
    if string.startswith('abj:'):
        return lilypondparsertools.parse_reduced_ly_syntax(string[4:])
    elif string.startswith('rtm:'):
        return rhythmtreetools.parse_rtm_syntax(string[4:])
    if language not in _lilypond_parsers_by_language:
        parser = lilypondparsertools.LilyPondParser(default_language=language)
        _lilypond_parsers_by_language[language] = parser
    return _lilypond_parsers_by_language[language](string)



# -*- coding: utf-8 -*-
import copy
from abjad.tools.abctools.AbjadValueObject import AbjadValueObject
from abjad.tools.topleveltools import new


class Annotation(AbjadValueObject):
    r'''Annotation.

    ..  container:: example

        ::

            >>> staff = Staff("c'8 d'8 e'8 f'8")
            >>> pitch = NamedPitch('ds')
            >>> annotation = indicatortools.Annotation('modifier', pitch)
            >>> attach(annotation, staff[0])
            >>> show(staff) # doctest: +SKIP

        ..  doctest::

            >>> print(format(staff))
            \new Staff {
                c'8
                d'8
                e'8
                f'8
            }

    Annotations contribute no formatting.
    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        '_default_scope',
        '_name',
        '_value',
        )

    ### INITIALIZER ###

    def __init__(self, name='annotation', value=None):
        self._default_scope = None
        if isinstance(name, type(self)):
            argument = name
            name = argument.name
            value = value or argument.value
        name = copy.copy(name)
        value = copy.copy(value)
        self._name = name
        self._value = value

    ### PUBLIC PROPERTIES ###

    @property
    def default_scope(self):
        r'''Gets default scope of annotation.

        ..  container:: example

            ::

                >>> pitch = NamedPitch('ds')
                >>> annotation = indicatortools.Annotation('modifier', pitch)
                >>> annotation.default_scope is None
                True

        Annotations are not scoped.

        Returns none.
        '''
        return self._default_scope

    @property
    def name(self):
        r'''Gets name of annotation.

        ..  container:: example

            ::

                >>> pitch = NamedPitch('ds')
                >>> annotation = indicatortools.Annotation('modifier', pitch)
                >>> annotation.name
                'modifier'

        Returns string.
        '''
        return self._name

    @property
    def value(self):
        r'''Gest value of annotation.

        ..  container:: example

            ::

                >>> pitch = NamedPitch('ds')
                >>> annotation = indicatortools.Annotation('modifier', pitch)
                >>> annotation.value
                NamedPitch('ds')

        Returns arbitrary object.
        '''
        return self._value



# -*- coding: utf-8 -*-
from abjad import *
from abjad.tools.pitchtools import NumberedPitch


def test_pitchtools_NumberedPitch___repr___01():
    r'''Numbered pitch repr is evaluable.
    '''

    numbered_pitch_1 = pitchtools.NumberedPitch(13)
    numbered_pitch_2 = eval(repr(numbered_pitch_1))

    assert isinstance(numbered_pitch_1, pitchtools.NumberedPitch)
    assert isinstance(numbered_pitch_2, pitchtools.NumberedPitch)



# -*- coding: utf-8 -*-
from abjad import *
from abjad.tools import tonalanalysistools


def test_tonalanalysistools_Scale_scale_degrees_01():

    scale = tonalanalysistools.Scale('g', 'major')

    assert scale.tonic == pitchtools.NamedPitchClass('g')
    assert scale.superdominant == pitchtools.NamedPitchClass('a')
    assert scale.mediant == pitchtools.NamedPitchClass('b')
    assert scale.subdominant == pitchtools.NamedPitchClass('c')
    assert scale.dominant == pitchtools.NamedPitchClass('d')
    assert scale.submediant == pitchtools.NamedPitchClass('e')
    assert scale.leading_tone == pitchtools.NamedPitchClass('fs')



# -*- coding: utf-8 -*-
from abjad.tools import durationtools
from abjad.tools import systemtools
from abjad.tools.scoretools.Tuplet import Tuplet
from abjad.tools.topleveltools import iterate
from abjad.tools.topleveltools import mutate


class FixedDurationTuplet(Tuplet):
    r'''A tuplet with fixed duration and variable multiplier.

    ..  container:: example

            >>> tuplet = scoretools.FixedDurationTuplet(Duration(2, 8), [])
            >>> tuplet.extend("c'8 d'8 e'8")
            >>> show(tuplet) # doctest: +SKIP

        ..  doctest::

            >>> print(format(tuplet))
            \times 2/3 {
                c'8
                d'8
                e'8
            }

        ::

            >>> tuplet.append("fs'4")
            >>> show(tuplet) # doctest: +SKIP

        ..  doctest::

            >>> print(format(tuplet))
            \times 2/5 {
                c'8
                d'8
                e'8
                fs'4
            }

    '''

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Containers'

    __slots__ = (
        '_target_duration',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        duration=durationtools.Duration(1, 4),
        music="c'8 c'8 c'8",
        ):
        dummy_multiplier = durationtools.Multiplier(1)
        Tuplet.__init__(self, dummy_multiplier, music)
        self._signifier = '@'
        self.target_duration = duration

    ### SPECIAL METHODS ###

    def __getnewargs__(self):
        '''Gets new arguments of fixed-duration tuplet.

        Returns tuple.
        '''
        duration = self.target_duration
        empty_components = ()
        return (duration, empty_components)

    def __repr__(self):
        '''Gets interpreter representation of fixed-duration tuplet.

        Returns string.
        '''
        result = '{}({!r}, {!r})'
        result = result.format(
            type(self).__name__,
            self.target_duration,
            self._contents_summary,
            )
        return result

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return systemtools.FormatSpecification(
            client=self,
            repr_args_values=[self.target_duration, self._contents_summary],
            storage_format_args_values=[self.target_duration, self[:]],
            storage_format_kwargs_names=[],
            )

    def _scale(self, multiplier):
        from abjad.tools import scoretools
        multiplier = durationtools.Multiplier(multiplier)
        # find new target duration
        old_target_duration = self.target_duration
        new_target_duration = multiplier * old_target_duration
        # change tuplet target duration
        self.target_duration = new_target_duration
        # if multiplier is note-head assignable,
        # scale contents graphically
        if multiplier.is_assignable:
            for component in self[:]:
                if isinstance(component, scoretools.Leaf):
                    new_duration = multiplier * component.written_duration
                    component._set_duration(new_duration)
        self._fix()

    ### PUBLIC METHODS ###

    def to_fixed_multiplier(self):
        r'''Changes fixed-duration tuplet to (unqualified) tuplet.

        ..  container:: example

            ::

                >>> tuplet = scoretools.FixedDurationTuplet((2, 8), [])
                >>> tuplet.extend("c'8 d'8 e'8")
                >>> show(tuplet) # doctest: +SKIP

            ::

                >>> tuplet
                FixedDurationTuplet(Duration(1, 4), "c'8 d'8 e'8")

            ::

                >>> new_tuplet = tuplet.to_fixed_multiplier()
                >>> show(new_tuplet) # doctest: +SKIP

            ::

                >>> new_tuplet
                Tuplet(Multiplier(2, 3), "c'8 d'8 e'8")

        Returns new tuplet.
        '''
        from abjad.tools import scoretools
        new_tuplet = scoretools.Tuplet(self.multiplier, [])
        mutate(self).swap(new_tuplet)
        return new_tuplet

    def toggle_prolation(self):
        r'''Toggles prolation of fixed-duration tuplet.

        ::

            >>> tuplet = scoretools.FixedDurationTuplet((1, 4), "c'8 d'8 e'8")
            >>> show(tuplet) # doctest: +SKIP

        ::

            >>> tuplet.toggle_prolation()
            >>> show(tuplet) # doctest: +SKIP

        ::

            >>> tuplet.toggle_prolation()
            >>> show(tuplet) # doctest: +SKIP

        Returns none.
        '''
        if self.is_diminution:
            while self.is_diminution:
                for leaf in iterate(self).by_leaf():
                    leaf.written_duration /= 2
        elif not self.is_diminution:
            while not self.is_diminution:
                for leaf in iterate(self).by_leaf():
                    leaf.written_duration *= 2

    def trim(self, start, stop='unused'):
        r'''Trims fixed-duration tuplet elements from `start` to `stop`:

        ::

            >>> tuplet = scoretools.FixedDurationTuplet(
            ...     Duration(2, 8), "c'8 d'8 e'8")
            >>> tuplet
            FixedDurationTuplet(Duration(1, 4), "c'8 d'8 e'8")

        ::

            >>> tuplet.trim(2)
            >>> tuplet
            FixedDurationTuplet(Duration(1, 6), "c'8 d'8")

        Preserves fixed-duration tuplet multiplier.

        Adjusts fixed-duration tuplet duration.

        Returns none.
        '''
        if stop != 'unused':
            assert not (start == 0 and (stop is None or len(self) <= stop))
        old_multiplier = self.multiplier
        if stop == 'unused':
            del(self[start])
        else:
            del(self[start:stop])
        self.target_duration = old_multiplier * self._contents_duration

    ### PUBLIC PROPERTIES ###

    @property
    def multiplied_duration(self):
        r'''Gets multiplied duration of tuplet.

        ::

            >>> tuplet = scoretools.FixedDurationTuplet((1, 4), "c'8 d'8 e'8")
            >>> tuplet.multiplied_duration
            Duration(1, 4)

        Returns duration.
        '''
        return self.target_duration

    @property
    def multiplier(self):
        r'''Gets and sets multiplier of fixed-duration tuplet.

        ::

            >>> tuplet = scoretools.FixedDurationTuplet(
            ...     (1, 4), "c'8 d'8 e'8")
            >>> tuplet.multiplier
            Multiplier(2, 3)

        Returns multiplier.
        '''
        if 0 < len(self):
            return durationtools.Multiplier(
                self.target_duration / self._contents_duration)
        else:
            return None

    @multiplier.setter
    def multiplier(self, argument):
        pass

    @property
    def target_duration(self):
        r'''Gets and sets target duration of fixed-duration tuplet.

        ::

            >>> tuplet = scoretools.FixedDurationTuplet(
            ...     (1, 4), "c'8 d'8 e'8")
            >>> tuplet.target_duration
            Duration(1, 4)

        ..  doctest::

            >>> print(format(tuplet))
            \times 2/3 {
                c'8
                d'8
                e'8
            }

        ::

            >>> tuplet.target_duration = Duration(5, 8)
            >>> print(format(tuplet))
            \tweak text #tuplet-number::calc-fraction-text
            \times 5/3 {
                c'8
                d'8
                e'8
            }

        Returns duration.
        '''
        return self._target_duration

    @target_duration.setter
    def target_duration(self, argument):
        target_duration = durationtools.Duration(argument)
        assert 0 < target_duration
        self._target_duration = target_duration



# -*- coding: utf-8 -*-
from abjad.tools import selectiontools


def make_repeated_notes_from_time_signature(time_signature, pitch="c'"):
    '''Make repeated notes from `time_signature`:

    ::

        >>> scoretools.make_repeated_notes_from_time_signature((5, 32))
        Selection([Note("c'32"), Note("c'32"), Note("c'32"), Note("c'32"), Note("c'32")])

    Make repeated notes with `pitch` from `time_signature`:

    ::

        >>> scoretools.make_repeated_notes_from_time_signature((5, 32), pitch="d''")
        Selection([Note("d''32"), Note("d''32"), Note("d''32"), Note("d''32"), Note("d''32")])

    Returns list of notes.
    '''
    from abjad.tools import indicatortools
    from abjad.tools import scoretools

    # afford basic input polymorphism
    time_signature = indicatortools.TimeSignature(time_signature)

    # check input
    if time_signature.has_non_power_of_two_denominator:
        raise NotImplementedError(
            'TODO: extend this function for time signatures with a non-power-of-two denominators.')

    # make and return repeated notes
    duration = (1, time_signature.denominator)
    result = time_signature.numerator * scoretools.Note(pitch, duration)

    # return result
    result = selectiontools.Selection(result)
    return result



# -*- coding: utf-8 -*-
from abjad.tools.datastructuretools.TypedTuple import TypedTuple


class PatternList(TypedTuple):
    r'''Pattern list.

    ..  container:: example

        Three patterns:

        ::

            >>> patterns = patterntools.PatternList([
            ...     patterntools.Pattern(
            ...         indices=[0, 1, 7],
            ...         period=10,
            ...         ),
            ...     patterntools.Pattern(
            ...         indices=[-2, -1],
            ...         ),
            ...     patterntools.Pattern(
            ...         indices=[2],
            ...         period=3,
            ...         ),
            ...     ])

        ::

            >>> f(patterns)
            patterntools.PatternList(
                (
                    patterntools.Pattern(
                        indices=[0, 1, 7],
                        period=10,
                        ),
                    patterntools.Pattern(
                        indices=[-2, -1],
                        ),
                    patterntools.Pattern(
                        indices=[2],
                        period=3,
                        ),
                    )
                )

    ..  container:: example

        Two patterns:

        ::

            >>> patterns = patterntools.PatternList([
            ...     patterntools.Pattern(
            ...         indices=[1],
            ...         period=2,
            ...         ),
            ...     patterntools.Pattern(
            ...         indices=[-3, -2, -1],
            ...         ),
            ...     ])

        ::

            >>> f(patterns)
            patterntools.PatternList(
                (
                    patterntools.Pattern(
                        indices=[1],
                        period=2,
                        ),
                    patterntools.Pattern(
                        indices=[-3, -2, -1],
                        ),
                    )
                )

    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        )

    ### PUBLIC METHODS ###

    def get_matching_pattern(self, index, total_length, rotation=None):
        r'''Gets pattern matching `index`.

        ..  container:: example

            Two patterns:

            ::

                >>> patterns = patterntools.PatternList([
                ...     patterntools.Pattern(
                ...         indices=[1],
                ...         period=2,
                ...         ),
                ...     patterntools.Pattern(
                ...         indices=[-3, -2, -1],
                ...         ),
                ...     ])

            Gets patterns that match the first ten indices:

            ::

                >>> for i in range(10):
                ...     match = patterns.get_matching_pattern(i, 10)
                ...     print(i, match)
                ...
                0 None
                1 Pattern(indices=[1], period=2)
                2 None
                3 Pattern(indices=[1], period=2)
                4 None
                5 Pattern(indices=[1], period=2)
                6 None
                7 Pattern(indices=[-3, -2, -1])
                8 Pattern(indices=[-3, -2, -1])
                9 Pattern(indices=[-3, -2, -1])

            Last three indices match the second pattern.

            Gets patterns that match next ten indices:

            ::

                >>> for i in range(10, 20):
                ...     match = patterns.get_matching_pattern(i, 10)
                ...     print(i, match)
                ...
                10 None
                11 Pattern(indices=[1], period=2)
                12 None
                13 Pattern(indices=[1], period=2)
                14 None
                15 Pattern(indices=[1], period=2)
                16 None
                17 Pattern(indices=[1], period=2)
                18 None
                19 Pattern(indices=[1], period=2)

            Last three indices no longer match the second pattern.

        ..  container:: example

            Gets patterns that match the first ten indices, with rotation set
            to ``1``:

            ::

                >>> for i in range(10):
                ...     match = patterns.get_matching_pattern(i, 10, rotation=1)
                ...     print(i, match)
                ...
                0 Pattern(indices=[1], period=2)
                1 None
                2 Pattern(indices=[1], period=2)
                3 None
                4 Pattern(indices=[1], period=2)
                5 None
                6 Pattern(indices=[1], period=2)
                7 Pattern(indices=[-3, -2, -1])
                8 Pattern(indices=[-3, -2, -1])
                9 Pattern(indices=[-3, -2, -1])

            Matching indices of first pattern offset by ``1``.

            Gets patterns that match next ten indices with rotation set to
            ``1``:

            ::

                >>> for i in range(10, 20):
                ...     match = patterns.get_matching_pattern(i, 10, rotation=1)
                ...     print(i, match)
                ...
                10 Pattern(indices=[1], period=2)
                11 None
                12 Pattern(indices=[1], period=2)
                13 None
                14 Pattern(indices=[1], period=2)
                15 None
                16 Pattern(indices=[1], period=2)
                17 None
                18 Pattern(indices=[1], period=2)
                19 None

            Matching indices of first pattern offset by ``1``.

        ..  container:: example

            With inverted patterns:

            ::

                >>> patterns = patterntools.PatternList([
                ...     patterntools.Pattern(
                ...         indices=[-3],
                ...         inverted=True,
                ...         ),
                ...     ])

            ::

                >>> for i in range(10):
                ...     match = patterns.get_matching_pattern(i, 10)
                ...     print(i, match)
                ...
                0 Pattern(indices=[-3], inverted=True)
                1 Pattern(indices=[-3], inverted=True)
                2 Pattern(indices=[-3], inverted=True)
                3 Pattern(indices=[-3], inverted=True)
                4 Pattern(indices=[-3], inverted=True)
                5 Pattern(indices=[-3], inverted=True)
                6 Pattern(indices=[-3], inverted=True)
                7 None
                8 Pattern(indices=[-3], inverted=True)
                9 Pattern(indices=[-3], inverted=True)

        Returns pattern or none.
        '''
        from abjad.tools import rhythmmakertools
        prototype = (
            rhythmmakertools.SilenceMask,
            rhythmmakertools.SustainMask,
            )
        for pattern in reversed(self):
            if isinstance(pattern, prototype):
                if pattern.pattern.matches_index(
                    index, total_length, rotation=rotation):
                    return pattern
            elif pattern.matches_index(index, total_length, rotation=rotation):
                return pattern

    def get_matching_payload(self, index, total_length, rotation=None):
        r'''Gets payload attached to pattern matching `index`.

        ..  container:: example

            Two patterns. Underlying notes with even divisions
            assigned to the last three indices:

            ::

                >>> patterns = patterntools.PatternList([
                ...     patterntools.Pattern(
                ...         indices=[0],
                ...         payload=rhythmmakertools.NoteRhythmMaker(),
                ...         period=1,
                ...         ),
                ...     patterntools.Pattern(
                ...         indices=[-3, -2, -1],
                ...         payload=rhythmmakertools.EvenDivisionRhythmMaker(),
                ...         ),
                ...     ])

            Over ten indices:

            ::

                >>> for i in range(10):
                ...     match = patterns.get_matching_payload(i, 10)
                ...     print(i, match)
                ...
                0 NoteRhythmMaker()
                1 NoteRhythmMaker()
                2 NoteRhythmMaker()
                3 NoteRhythmMaker()
                4 NoteRhythmMaker()
                5 NoteRhythmMaker()
                6 NoteRhythmMaker()
                7 EvenDivisionRhythmMaker(denominators=[8], preferred_denominator='from_counts')
                8 EvenDivisionRhythmMaker(denominators=[8], preferred_denominator='from_counts')
                9 EvenDivisionRhythmMaker(denominators=[8], preferred_denominator='from_counts')

            Over fifteen indices:

            ::

                >>> for i in range(15):
                ...     match = patterns.get_matching_payload(i, 15)
                ...     print(i, match)
                ...
                0 NoteRhythmMaker()
                1 NoteRhythmMaker()
                2 NoteRhythmMaker()
                3 NoteRhythmMaker()
                4 NoteRhythmMaker()
                5 NoteRhythmMaker()
                6 NoteRhythmMaker()
                7 NoteRhythmMaker()
                8 NoteRhythmMaker()
                9 NoteRhythmMaker()
                10 NoteRhythmMaker()
                11 NoteRhythmMaker()
                12 EvenDivisionRhythmMaker(denominators=[8], preferred_denominator='from_counts')
                13 EvenDivisionRhythmMaker(denominators=[8], preferred_denominator='from_counts')
                14 EvenDivisionRhythmMaker(denominators=[8], preferred_denominator='from_counts')

        '''
        pattern = self.get_matching_pattern(index, total_length, rotation=rotation)
        payload = None
        if pattern:
            payload = pattern.payload
        return payload

    ### PRIVATE PROPERTIES ###

    @property
    def _item_coercer(self):
        from abjad.tools import patterntools
        from abjad.tools import rhythmmakertools
        prototype = (
            rhythmmakertools.SilenceMask,
            rhythmmakertools.SustainMask,
            )
        def coerce_(argument):
            if isinstance(argument, prototype):
                pass
            elif not isinstance(argument, patterntools.Pattern):
                argument = patterntools.Pattern(*argument)
            return argument
        return coerce_
